import { createNoise3D } from "simplex-noise";

export const generateGlobeTextureData = (width: number, height: number, seed?: number): {
    data: Uint8Array;
    width: number;
    height: number;
} => {
    const size = width * height;
    const data = new Uint8Array(4 * size); // RGBA
    seed = seed ?? Math.floor(Math.random() * 10000); // 시드 초기화 (랜덤 시드 생성)
    const getRandom = (seed: number) => () => {
        // 시드가 주어지면 고정된 랜덤 생성기 사용
        let x = Math.sin(seed) * 10000;
        x = (x - Math.floor(x)) * 10000;
        seed += 1; // 시드 증가
        return x - Math.floor(x); // 0 ~ 1 사이 값 반환
    } // 시드가 주어지면 고정된 랜덤 생성기 사용 (선택 사항)
    const noise3D = createNoise3D(getRandom(seed)); // 3D 노이즈 생성 함수
    const noise3D2 = createNoise3D(getRandom(seed ^ 12)); // 3D 노이즈 생성 함수 (다른 스케일을 위해)
    const noise3D3 = createNoise3D(getRandom(seed ^ 224)); // 3D 노이즈 생성 함수 (다른 스케일을 위해)

    const noiseScale = 2; // 노이즈 스케일 (값이 클수록 지형이 작아짐)
    const waterLevel = 0.05; // 해수면 높이 (-1 ~ 1 사이 값)
    const mountainLevel = 0.4; // 산악 지형 시작 높이
    const snowLine = 0.8; // 만년설 시작 높이 (위도와 조합)

    for (let i = 0; i < size; i++) {
        const stride = i * 4;
        const u = (i % width) / width; // 0 to 1 (경도 방향)
        const v = Math.floor(i / width) / height; // 0 to 1 (위도 방향)

        // 위도, 경도를 3D 좌표로 변환 (단위 구 기준)
        const lat = (v - 0.5) * 180;
        const lon = (u - 0.5) * 360;
        const phi = (90 - lat) * (Math.PI / 180);
        const theta = (lon + 180) * (Math.PI / 180);

        const x = -Math.sin(phi) * Math.cos(theta);
        const z = Math.sin(phi) * Math.sin(theta);
        const y = Math.cos(phi);

        // 3D Simplex Noise 값 계산 (-1 ~ 1)
        const noiseValue = noise3D(x * noiseScale, y * noiseScale, z * noiseScale)
        + noise3D2(x * noiseScale * 0.5, y * noiseScale * 0.5, z * noiseScale * 0.5) * 0.5
        + noise3D3(x * noiseScale * 0.25, y * noiseScale * 0.25, z * noiseScale * 0.25) * 0.25; // 노이즈 조합

        let r = 50; // 바다 기본 R
        let g = 100; // 바다 기본 G
        let b = 200; // 바다 기본 B
        let a = 255; // Alpha

        const absLat = Math.abs(lat);

        if (noiseValue > waterLevel) {
            // 육지
            const elevation = (noiseValue - waterLevel) / (1 - waterLevel); // 0 ~ 1 정규화

            if (elevation < mountainLevel) {
                // 평지/언덕 (녹색)
                r = 80 + 70 * elevation + Math.random() * 10;
                g = 140 + 60 * (1 - elevation) + Math.random() * 10;
                b = 40 + 30 * elevation + Math.random() * 5;
            } else {
                // 산악 (갈색/회색)
                const mountainFactor = (elevation - mountainLevel) / (1 - mountainLevel); // 0 ~ 1
                r = 140 + 40 * mountainFactor;
                g = 120 + 20 * mountainFactor;
                b = 100 + 10 * mountainFactor;

                // 높은 산 + 높은 위도 = 눈 (흰색)
                const snowChance = Math.max(0, (absLat - 60) / 30) + mountainFactor * 0.5; // 위도 60도 이상부터 눈 확률 증가
                if (elevation > snowLine || snowChance > 0.5) {
                    r = 240;
                    g = 240;
                    b = 245;
                }
            }
        } else {
            // 바다 (깊이에 따라 색상 변화)
            const depthFactor = Math.abs((noiseValue - waterLevel) / ( 1 + waterLevel)); // 0 ~ 1 정규화 (깊이 비율)
            r = 50 - 30 * depthFactor;
            g = 100 - 50 * depthFactor;
            b = 200 - 40 * depthFactor;
        }

        // 극지방 얼음 (기존 로직과 유사하게, 노이즈 값과 무관하게 덮어쓰기)
        if (absLat > 75) {
            let polarFactor = (absLat - 75) / 15; // 0 to 1
            polarFactor = 1 - Math.pow(1 - polarFactor, 5);

            // 얼음/눈 (흰색)
            r = 245 * polarFactor + (1 - polarFactor) * r;
            g = 245 * polarFactor + (1 - polarFactor) * g;
            b = 250 * polarFactor + (1 - polarFactor) * b;

        }


        data[stride] = Math.max(0, Math.min(255, r));
        data[stride + 1] = Math.max(0, Math.min(255, g));
        data[stride + 2] = Math.max(0, Math.min(255, b));
        data[stride + 3] = a; // Alpha 값 설정
    }

       // 데이터 배열과 크기를 반환
       return { data, width, height };
};
